from collections import OrderedDict
from amolkit.stringmanip import getPenaltyfromString

def readCharmmParameter(parfile):
    """
    Get Charmm parameters from any file containing parameters.
    Arguments:
        parfile: It can be any file format containing charmm parameters in correct style.
    Returns:
        parameters: Dictionary with keywords BONDS, ANGLES, DIHEDRALS and IMPROPERS
                    Example: parameter["BONDS"] = [[attyp1,attyp2,kval,eqdist],...]
                    Example: parameter["ANGLES"] = [[attyp1,attyp2,attyp3,kval,eqangl],...]
                    Example: parameter["DIHEDRALS"] = [[attyp1,attyp2,attyp3,attyp4,kval,mult,eqangl],...]
    Note:
        Have to add extraction of vdw parameters.
        Canonicalization of parameters need to be performed from a separate function. 
        Because user may be reading multiple files using this function. 
        User should latter canonicalize and only keep unique parameters.
    """

    parameters={}
    for marker in ["BONDS","ANGLES","DIHEDRALS","IMPROPERS"]:
        parameters[marker] = [] 

    marker = ""
    with open(parfile,'r') as readpar:
        for line in readpar:
            field = line.strip().split("!")[0].split()
            comment = "!".join(line.strip().split("!")[1:])
            if len(field) > 0:
                if len(field) == 1:
                    if field[0].upper() == "BONDS":
                        marker = "BONDS"
                    if field[0].upper() == "ANGLES":
                        marker = "ANGLES"
                    if field[0].upper() == "DIHEDRALS":
                        marker = "DIHEDRALS"
                    if field[0].upper() == "IMPROPER":
                        marker = "IMPROPERS"
                try:
                    if marker in ["BONDS","ANGLES","DIHEDRALS","IMPROPERS"]:
                        penalty,comment = getPenaltyfromString(comment)
                    if marker == "BONDS" and len(field) == 4:
                        parameters[marker].append([field[0],field[1],round(float(field[2]),4),round(float(field[3]),4),penalty,comment])
                    if marker == "ANGLES" and len(field) == 5:
                        parameters[marker].append([field[0],field[1],field[2],round(float(field[3]),4),round(float(field[4]),4),penalty,comment])
                    if marker == "ANGLES" and len(field) == 7:
                        parameters[marker].append([field[0],field[1],field[2],round(float(field[3]),4),round(float(field[4]),4),round(float(field[5]),4),round(float(field[6]),4),penalty,comment])
                    if marker in ["DIHEDRALS","IMPROPERS"] and len(field) == 7:
                        parameters[marker].append([field[0],field[1],field[2],field[3],round(float(field[4]),4),int(field[5]),round(float(field[6]),1),penalty,comment])
                except (ValueError,IndexError):
                    pass
    return parameters

def readCharmmParameterFiles(parfiles:list):
    """
    Get Charmm parameters from multiple files and collect in parameters dictionary
    Arguments:
       parfiles: list of files containing charmm parameters
    Returns:
       parameters: Dictionary containing all parameters obtained from list in parfiles
    """

    if isinstance(parfiles,str):
        parfiles=[parfiles]

    parameters={}
    for marker in ["BONDS","ANGLES","DIHEDRALS","IMPROPERS"]:
        parameters[marker] = [] 

    for parfile in parfiles:
        cparams = readCharmmParameter(parfile)
        for key,value in cparams.items():
            if not parameters[key]: 
               parameters[key] = cparams[key]
            else:
               parameters[key].append(cparams[key])
    return parameters           

def canonical_format(iclist,icnumlist):
    # Horizontal swapping using ranking scheme
    for i in range(len(iclist)):
        if len(iclist[i]) == 2:
           ric1,ric2= ranking(iclist[i][0],iclist[i][1])
           if ric1 > ric2:
              iclist[i] = iclist[i][::-1]
              icnumlist[i] = icnumlist[i][::-1]
              
        if len(iclist[i]) == 3:
           ric1,ric2 = ranking(iclist[i][0],iclist[i][2])
           if ric1 > ric2:
              iclist[i] = iclist[i][::-1]
              icnumlist[i] = icnumlist[i][::-1]

        if len(iclist[i]) == 4:
           ric1,ric2 = ranking(iclist[i][0],iclist[i][3])
           if ric1 > ric2:
              iclist[i] = iclist[i][::-1]
              icnumlist[i] = icnumlist[i][::-1]
              
    # Vertical swapping using alphabetical order
    iclist = [i for i in sorted(iclist, key=lambda ic:ic[0])]
    icnumlist = [i for _,i in sorted(zip(iclist,icnumlist), key=lambda ic:ic[0][0])]
    return(iclist,icnumlist)

def ranking(atnm1,atnm2):
    # If in future new elements are added to rank dict then forced ranking 11 and 12 
    # should change accordingly.
    rank={"C":1,"N":2,"O":3,"P":4,"S":5,"F":6,"CL":7,"BR":8,"B":9,"AL":10,"H":30}
    ratnm1 = rank.get(atnm1[0:2].upper())
    if ratnm1 == None: ratnm1 = rank.get(atnm1[0:1].upper())
    ratnm2 = rank.get(atnm2[0:2].upper())
    if ratnm2 == None: ratnm2 = rank.get(atnm2[0:1].upper())
    if ratnm1 == None and ratnm2 == None:
        #if ratnm1 > ratnm2:
        ratnm1 = 11 
        ratnm2 = 12
        #else:
        #    ratnm1 = 12 
        #    ratnm2 = 11
    elif ratnm1 == None and ratnm2 != None:
        ratnm1 = 11
    elif ratnm1 != None and ratnm2 == None:
        ratnm2 = 11
    return (ratnm1,ratnm2)     

class Parameter():
    def __init__(self):
        """
        Constructor for the Parameter class.
        """

    def loadCharmmParameter(self,parfiles):
        """
        Load topology information using CHARMM force field and store it as attributes.

        Args:
            resname (str): Name of the residue.
            resitopfile (str): Path to the CHARMM topology file.
        """

        self.charmmParameter = readCharmmParameterFiles(parfiles)



import sys
import math
import argparse
from collections import OrderedDict
from amolkit import genic
from amolkit import getEleInfo as gei 
from amolkit.topology import Topology

def dcharge(alpha):
   kdrude = 500.0
   ccelec = 332.0716
   qdip = 2*kdrude*alpha/ccelec 
   qdip = -(math.sqrt(abs(qdip)))
   qdip = round(qdip,4)
   return (qdip)

def isotropy(qdip):
   kdrude = 500.0
   iso = 1.0/((qdip**2)/kdrude)
   return (iso)

def anisotropy(a11,a22):
   kdrude = 500.0
   a33 = 3 - a11 - a22
   k11 = 1/a11
   k22 = 1/a22
   k33 = 1/a33 
   k11 = kdrude*k11
   k22 = kdrude*k22
   k33 = kdrude*k33
   k33 = k33 - kdrude
   k11 = k11 - kdrude - k33
   k22 = k22 - kdrude - k33
   return ([k11,k22,k33])

class Psf():
    def __init__ (self,topology): 
        """
        Constructor for Psf class
        Using topology to assign relevant attributes.
        """

        self.resname = None 
        self.segname = None
        self.first_byIdx=None
        self.last_byIdx=None
        self.npsfatoms=None

        self.atomindex_byId=OrderedDict() 
        self.atomtype_byId=OrderedDict()
        self.atomcharge_byId=OrderedDict()
        self.atommass_byId=OrderedDict()
        self.atomele_byId=OrderedDict()
        self.atomalpha_byId=OrderedDict()
        self.atomthole_byId=OrderedDict()
        self.atomsinring=OrderedDict()

        self.atomserial_byIdx=OrderedDict()
        self.atomresn_byIdx=OrderedDict()
        self.atomresid_byIdx=OrderedDict()
        self.atomsegn_byIdx=OrderedDict()
        self.atomname_byIdx=OrderedDict()
        self.atomtype_byIdx=OrderedDict()
        self.atomcharge_byIdx=OrderedDict()
        self.atommass_byIdx=OrderedDict()
        self.atomalpha_byIdx=OrderedDict()
        self.atomthole_byIdx=OrderedDict()
        self.atomqlp_byIdx=OrderedDict() 
        self.groups=[]
        self.bondindices=[]
        self.lpbondindices=[]
        self.drudebondindices=[]
        self.angleindices=[]
        self.dihedralindices=[]
        self.improperindices=[]
        self.donorindices=[]
        self.acceptorindices=[]
        self.cmapindices=[]
        self.lpics=[]
        self.anisotropies=[]
        self.drudebondindices=[] 
        self.drudebonds=[] 

        self.topology = topology 

    def genpsf(self,istart=0,resid=1,dweight=0.4,segname=None,autogenangdih=True): 
        '''
        Load the features of the residue based on topology file.
        TODO: read the topology file separately and get all features from there.
        '''

        atomcharges=[]

        resname = self.topology.resname
        segname = segname if segname else resname
        self.fftype = self.topology.fftype
        self.atomsinring = self.topology.atomsinring 
        self.groups = self.topology.groups 

        nat=istart
        if self.fftype[0:4] == "addi":
            for key,value in self.topology.atomname_byIdx.items():
                # Rethink: nat should be related to key # Although this would work because atomname_byIdx is ordereddict
                nat = nat + 1 
                self.atomname_byIdx[nat]   = value
                self.atomindex_byId[value] = nat
                self.atomserial_byIdx[nat] = nat
                self.atomtype_byIdx[nat]   = self.topology.atomtype_byId[value] 
                self.atomqlp_byIdx[nat]    = -1 if self.atomtype_byIdx[nat][0:2].upper() == "LP" else 0            
                self.atomcharge_byIdx[nat] = self.topology.atomcharge_byId[value] 
                self.atommass_byIdx[nat]   = self.topology.atommass_byId[value] 
                self.atomresn_byIdx[nat]   = resname
                self.atomresid_byIdx[nat]  = resid 
                self.atomsegn_byIdx[nat]   = segname 
                atomcharges.append(self.atomcharge_byIdx[nat])

        elif self.fftype[0:4] == "drud":
            for key,value in self.topology.atomname_byIdx.items():
                # Rethink: nat should be related to key # Although this would work because atomname_byIdx is ordereddict
                nat = nat + 1 
                # Only has list of drude containing atoms
                drud = self.topology.atomdrudetype_byId.get(value)
                self.atomname_byIdx[nat]   = value
                self.atomindex_byId[value] = nat
                self.atomserial_byIdx[nat] = nat
                self.atomtype_byIdx[nat]   = self.topology.atomtype_byId[value] 
                self.atomqlp_byIdx[nat]    = -1 if self.atomtype_byIdx[nat][0:2].upper() == "LP" else 0            
                self.atomcharge_byIdx[nat] = self.topology.atomcharge_byId[value] 
                self.atomalpha_byIdx[nat]  = self.topology.atomalpha_byId[value] if drud else 0.0 
                self.atomthole_byIdx[nat]  = self.topology.atomthole_byId[value] if drud else 0.0
                self.atommass_byIdx[nat]   = self.topology.atommass_byId[value] 
                self.atomresn_byIdx[nat]   = resname
                self.atomresid_byIdx[nat]  = resid 
                self.atomsegn_byIdx[nat]   = segname 
                atomcharges.append(self.atomcharge_byIdx[nat])

                if drud:
                    nat = nat + 1
                    dvalue = "D"+value 
                    self.atomname_byIdx[nat]       = dvalue
                    self.atomindex_byId[dvalue]    = nat
                    self.atomserial_byIdx[nat]     = nat
                    self.atomtype_byIdx[nat]       = drud 
                    self.atomqlp_byIdx[nat]        = 0            
                    self.atomcharge_byIdx[nat]     = dcharge(self.topology.atomalpha_byId[value]) 
                    self.atomcharge_byIdx[nat-1]   = self.atomcharge_byIdx[nat-1] - self.atomcharge_byIdx[nat]
                    self.atomalpha_byIdx[nat]      = 0.0 
                    self.atomthole_byIdx[nat]      = 0.0 
                    self.atommass_byIdx[nat]       = dweight
                    self.atommass_byIdx[nat-1]     = self.atommass_byIdx[nat-1] - self.atommass_byIdx[nat]
                    self.atomresn_byIdx[nat]       = resname
                    self.atomresid_byIdx[nat]      = resid 
                    self.atomsegn_byIdx[nat]       = segname 
                    atomcharges.append(0.0)
                    self.drudebondindices.append([nat,nat-1])
                    self.drudebonds.append([value,dvalue]) 

        for ba,bb in self.topology.bonds:
            self.bondindices.append([self.atomindex_byId[ba],self.atomindex_byId[bb]])

        for ba,bb in self.topology.lpbonds:
            self.lpbondindices.append([self.atomindex_byId[ba],self.atomindex_byId[bb]])

        if self.topology.first_byId: self.first_byIdx = self.atomindex_byId[self.topology.first_byId]
        if self.topology.last_byId: self.last_byIdx = self.atomindex_byId[self.topology.last_byId]

        if autogenangdih:
            self.bondgraph = genic.GetBondAngleDihedral.GetBondDict(self.atomname_byIdx,self.bondindices)["aNoLPBond"]
            _,self.angleindices = genic.GetBondAngleDihedral.GetAngles(self.atomindex_byId,self.bondgraph)
            _,self.dihedralindices = genic.GetBondAngleDihedral.GetDihedrals(self.atomindex_byId,self.bondgraph)

            #Checks if there are any atoms not bonded to anything or any bonds with no defined atoms
            for atom in list(self.bondgraph.keys()):
                if atom not in self.atomindex_byId.keys():
                     raise RuntimeError(('A bond is defined for atom %s but it is not in atom list of %s.' %(atom,resname)))

        else:
            for aa,ab,ac in self.topology.angles:
                self.angleindices.append([self.atomindex_byId[aa],self.atomindex_byId[ab],self.atomindex_byId[ac]])

            for da,db,dc,dd in self.topology.dihedrals:
                self.dihedralindices.append([self.atomindex_byId[da],self.atomindex_byId[db],self.atomindex_byId[dc],self.atomindex_byId[dd]])

        for ima,imb,imc,imd in self.topology.impropers:
            self.improperindices.append([self.atomindex_byId[ima],self.atomindex_byId[imb],self.atomindex_byId[imc],self.atomindex_byId[imd]])

        for da,db in self.topology.donors:
            self.donorindices.append([self.atomindex_byId[da],self.atomindex_byId[db]])

        for aa,ab in self.topology.acceptors:
            self.acceptorindices.append([self.atomindex_byId[aa],self.atomindex_byId[ab]])

        for ca,cb,cc,cd,ce,cf,cg,ch in self.topology.cmaps:
            self.cmapindices.append([self.atomindex_byId[ca],self.atomindex_byId[cb],self.atomindex_byId[cc],self.atomindex_byId[cd],self.atomindex_byId[ce],self.atomindex_byId[cf],self.atomindex_byId[cg],self.atomindex_byId[ch]])

        #print(self.topology.lpics)
        for i in range(len(self.topology.lpics)):
            hlist = list(map(lambda x:self.atomindex_byId[x],self.topology.lpics[i][1])) 
            self.lpics.append([self.topology.lpics[i][0],hlist,self.topology.lpics[i][2]])   

        for i in range(len(self.topology.anisotropies)):
            hlist = list(map(lambda x:self.atomindex_byId[x],self.topology.anisotropies[i][0]))
            vlist = anisotropy(*self.topology.anisotropies[i][1])
            self.anisotropies.append([hlist,vlist])

        self.npsfatoms=len(self.atomindex_byId.keys())
        # Determine group types
        for i in range(len(self.groups)):
            if i+1 <= len(self.groups): 
                gc = abs(sum(atomcharges[self.groups[i][0]:self.npsfatoms]))
                natgrp = self.npsfatoms - self.groups[i][0]
            else:
                gc = abs(sum(atomcharges[self.groups[i][0]:self.groups[i+1][0]]))
                natgrp = self.groups[i+1][0] - self.groups[i][0]
            if gc > 0.00001:
                self.groups[i][1] = 2
            else: 
                self.groups[i][1] = 1 if natgrp != 1 else 0
                
def psfFormat(fout,datatype,data):
    toprint =""
    counter = 0
    if datatype == "bonds":batch = 4
    if datatype == "angles":batch = 3
    if datatype == "dihedrals":batch = 2
    if datatype == "impropers":batch = 2
    if datatype == "donors":batch = 4
    if datatype == "acceptors":batch = 4
    if datatype == "nonbonds":batch = 8
    if datatype == "groups":batch = 3
    if datatype == "molnt":batch = 8
    if datatype == "lonepairs":batch = 2
    if datatype == "aniso":batch = 2
    if datatype == "cmap":batch = 2
    
    for i in range(len(data)):
        counter = counter + 1
        if counter%batch == 0:  
           topre   = "     ".join(map(lambda tp: str(tp).rjust(5),data[i]))
           toprint = "     ".join((toprint,topre))
           toprint = toprint+"\n"
        else:
           topre   = "     ".join(map(lambda tp: str(tp).rjust(5),data[i]))
           toprint = "     ".join((toprint,topre))
    fout.write("{:}\n".format(toprint))
    if len(data) == 0 or len(data)%batch !=0: fout.write("\n")
    return toprint

def writePsf(psf,fileout=None):
    import datetime
    from itertools import cycle 
    now = datetime.datetime.now()

    if fileout == None:
       fileout = "molecule.psf"
       print ("No output psffile name provided. Writing in molecule.psf") 

    allbonds         = [*psf.bondindices,*psf.drudebondindices]
    nallbonds        = len(allbonds)
    nallangles       = len(psf.angleindices)
    nalldihedrals    = len(psf.dihedralindices)
    nallimpropers    = len(psf.improperindices)
    nalldonors       = len(psf.donorindices)
    nallacceptors    = len(psf.acceptorindices)
    nallgroups       = len(psf.groups)
    nlpics           = len(psf.lpics)
    hostlpics        = [col[1] for col in psf.lpics]
    valulpics        = [col[2] for col in psf.lpics]
    nhostlpics       = sum(list(map(lambda x:len(x),hostlpics))) 
    nallanisotropies = len(psf.anisotropies)
    hostanisotropies = [row[0] for row in psf.anisotropies]
    valuanisotropies = [row[1] for row in psf.anisotropies]
    nallcmaps        = len(psf.cmapindices)

    fout = open(fileout,"w")
    if psf.fftype[0:4].lower() == "addi":
        fout.write("PSF EXT CMAP CHEQ XPLOR\n")
    elif psf.fftype[0:4].lower() == "drud":
        fout.write("PSF EXT CMAP CHEQ DRUDE XPLOR\n")
    fout.write("\n")
    fout.write('{:>10} !NTITLE\n'.format(2))
    fout.write("* Generated by amolkit\n")
    fout.write("* DATE:     "+now.strftime("%Y-%m-%d %H:%M")+"      WRITTEN BY: Anmol Kumar\n")
    fout.write("\n")
    fout.write('{:>10} !NATOM\n'.format(psf.npsfatoms))

    if psf.fftype[0:4].lower() == "addi": 
        ind = 0
        for i in range(psf.npsfatoms):
            ind = i + 1
            fout.write("{:>10} {:<8} {:<8} {:<8} {:<8} {:<8}{:>9.6f}    {:>10.4f}          {:>2}   0.00000     -0.301140E-02\n".format(psf.atomserial_byIdx[ind],
                psf.atomresn_byIdx[ind],psf.atomresid_byIdx[ind],
                psf.atomsegn_byIdx[ind],psf.atomname_byIdx[ind],
                psf.atomtype_byIdx[ind],psf.atomcharge_byIdx[ind],
                psf.atommass_byIdx[ind],psf.atomqlp_byIdx[ind]))

    elif psf.fftype[0:4].lower() == "drud": 
        ind = 0
        for i in range(psf.npsfatoms):
            ind = i + 1
            fout.write("{:>10} {:<8} {:<8} {:<8} {:<8} {:<8}{:>9.6f}    {:>10.4f}          {:>2}   {:>10.5f}    {:>10.5f}\n".format(psf.atomserial_byIdx[ind],
                psf.atomresn_byIdx[ind],psf.atomresid_byIdx[ind],
                psf.atomsegn_byIdx[ind],psf.atomname_byIdx[ind],
                psf.atomtype_byIdx[ind],psf.atomcharge_byIdx[ind],
                psf.atommass_byIdx[ind],psf.atomqlp_byIdx[ind],
                psf.atomalpha_byIdx[ind],psf.atomthole_byIdx[ind])) 

    fout.write("\n")

    fout.write('{:>10} !NBOND: bonds\n'.format(nallbonds))
    psfFormat(fout,"bonds",allbonds)
    fout.write('{:>10} !NTHETA: angles\n'.format(nallangles))
    psfFormat(fout,"angles",psf.angleindices)
    fout.write('{:>10} !NPHI: dihedrals\n'.format(nalldihedrals))
    psfFormat(fout,"dihedrals",psf.dihedralindices)
    fout.write('{:>10} !NIMPHI: impropers\n'.format(nallimpropers))
    psfFormat(fout,"impropers",psf.improperindices)
    fout.write('{:>10} !NDON: donors\n'.format(nalldonors))
    psfFormat(fout,"donors",psf.donorindices) 
    fout.write('{:>10} !NACC: acceptors\n'.format(nallacceptors))
    psfFormat(fout,"acceptors",psf.acceptorindices) 
    fout.write("{:>10} !NNB\n\n".format(0))
    psfFormat(fout,"nonbonds",[[0]]*psf.npsfatoms)
    fout.write("{:>10}         0 !NGRP NST2\n".format(nallgroups))
    psfFormat(fout,"groups",psf.groups)
    fout.write("{:>10} !MOLNT\n".format(1))
    psfFormat(fout,"molnt",[[1]]*psf.npsfatoms)
    fout.write("{:>10}{:>10} !NUMLP NUMLPH\n".format(nlpics,nhostlpics))

    lppos = 1
    for i in range(len(psf.lpics)):
        if psf.lpics[i][0].upper() in ["RELA","BISE"]:
            fout.write("{:>10}{:>10}   F {:>10.5f}   {:>10.4f}   {:>10.5f} \n".format(3,lppos,*psf.lpics[i][2]))
        elif psf.lpics[i][0].upper() in ["COLI"]:
            fout.write("{:>10}{:>10}   F {:>10.5f}   {:>10.4f}   {:>10.5f} \n".format(2,lppos,*psf.lpics[i][2]))
        lppos = lppos + len(psf.lpics[i][1])

    psfFormat(fout,"lonepairs",hostlpics) 

    if psf.fftype[0:4].lower() == "drud":
        fout.write("{:>10} !NUMANISO\n".format(nallanisotropies))
        for i in range(len(valuanisotropies)):
            fout.write("         {:>10.3f}   {:>10.4f}   {:>10.4f} \n".format(*valuanisotropies[i]))

        psfFormat(fout,"aniso",hostanisotropies)

    fout.write('{:>10} !NCRTERM: cross-terms\n'.format(nallcmaps))
    psfFormat(fout,"cmap",psf.cmapindices) 

    fout.close()
    prntscr = " ".join((str(fileout),"created"))
    print(prntscr) 

def buildPsf(resnames,resncount=None,resitopfiles=None,istart=0,resid=1,dweight=0.4,segname=None,autogenangdih=True,prevpsf=None): 
    if not resitopfiles:
        raise FileNotFoundError("No resi toppar file is provided.")

    if isinstance(resnames,str): resname=[resnames]
    if isinstance(resitopfiles,str): resitopfiles=[resitopfiles]
    
    if not resncount: resncount=[1]*len(resnames)
    psf=prevpsf
    first_byIdx=[]
    last_byIdx=[]
    restop=[]
    if not psf:
        t = Topology()
        for i,resn in enumerate(resnames):
            if i == 0: 
                istart = 0 
                resnwttopo = [resn]
            else: 
                if resn in resnwttopo:
                    t=resnwttopo.index(resn)
                else:
                    t.loadTopology(resn,resitopfiles)
                    restop.append(t)
                    resnwttopo.append(resn)
            p = Psf(t)
            p.genpsf(istart,resid,dweight,segname,autogenangdih)
            if p.first_byIdx: first_byIdx.append(p.first_byIdx)
            if p.last_byIdx: last_byIdx.append(p.last_byIdx)

            istart = istart + p.npsfatoms
            resid = resid + 1

            #if first and last:
            #    self.bondindices.append(istart,self.atomindex_byId[])



import os,sys
import amolkit 
from collections import OrderedDict

def test_readaddi():
    from amolkit.topology import readCharmmTopology
    
    allinfo=open("readaddi.txt","r").readlines()
    
    for e in allinfo:
        a=e.split()
        top=readCharmmTopology(a[0],a[1])
        assert top["rescharge"] == float(a[2])

def test_readdrud():
    from amolkit.topology import readCharmmTopology
    allinfo=open("readdrud.txt","r").readlines()
    
    for e in allinfo:
        a=e.split()
        top=readCharmmTopology(a[0],a[1])
        assert top["rescharge"] == float(a[2])

def test_genpsf_readaddi():
    from amolkit.topology import readCharmmTopology
    from amolkit.topology import Topology
    from amolkit.topology import Psf
    
    t=Topology()
    t.loadCharmmTopology("ACET","readaddi/top_all36_cgenff.rtf")
    p=Psf(t)
    p.genpsf()
    npsfatoms = 7
    print (p.npsfatoms)
    assert p.npsfatoms == npsfatoms

def test_genpsf_readdrud():
    from amolkit.topology import readCharmmTopology
    from amolkit.topology import Topology
    from amolkit.topology import Psf
    
    t=Topology()
    t.loadCharmmTopology("ALAD","readdrud/toppar_drude_master_protein_2019g.str")
    p=Psf(t)
    p.genpsf()
    npsfatoms = 36
    print (p.npsfatoms)
    assert p.npsfatoms == npsfatoms

def test_genpsf_mainaddi():
    from amolkit.topology import readCharmmTopology
    from amolkit.topology import Topology
    from amolkit.topology import Psf
    
    t=Topology()
    t.loadCharmmTopology("OXAZ","mainaddi/oxaz.str")
    p=Psf(t)
    p.genpsf()
    npsfatoms = 11
    print (p.npsfatoms)
    assert p.npsfatoms == npsfatoms

def test_genpsf_maindrud():
    from amolkit.topology import readCharmmTopology
    from amolkit.topology import Topology
    from amolkit.topology import Psf
    
    t=Topology()
    t.loadCharmmTopology("OXAZ","maindrud/oxaz.str")
    p=Psf(t)
    p.genpsf()
    npsfatoms = 11
    print (p.npsfatoms)
    assert p.npsfatoms == npsfatoms

def test_molsystem(tmp_path):
    from amolkit.molecule import Molecule
    from amolkit.molecule import writepdb
    from amolkit.molsystem import MolSystem
    
    resn=MolSystem()
    resn.loadTopology("OXAZ","mainaddi/oxaz.str")
    resn.loadParameters("mainaddi/oxaz.str")
    bondparam=[['C1', 'C2', None, None, None], ['C1', 'N', None, None, None], 
            ['C1', 'H1', None, None, None], ['C1', 'H2', None, None, None], 
            ['C2', 'O1', None, None, None], ['C2', 'H3', None, None, None], 
            ['C2', 'H4', None, None, None], ['O1', 'C3', None, None, None], 
            ['C3', 'O2', None, None, None], ['C3', 'N', None, None, None], 
            ['N', 'H5', None, None, None]]
    angleparam=[['C2', 'C1', 'N', None, None, None], ['C2', 'C1', 'H1', None, None, None], 
            ['C2', 'C1', 'H2', None, None, None], ['N', 'C1', 'H1', None, None, None], 
            ['N', 'C1', 'H2', None, None, None], ['H1', 'C1', 'H2', None, None, None], 
            ['C1', 'C2', 'O1', None, None, None], ['C1', 'C2', 'H3', None, None, None], 
            ['C1', 'C2', 'H4', None, None, None], ['O1', 'C2', 'H3', None, None, None], 
            ['O1', 'C2', 'H4', None, None, None], ['H3', 'C2', 'H4', None, None, None], 
            ['C2', 'O1', 'C3', None, None, None], ['O1', 'C3', 'O2', None, None, None], 
            ['O1', 'C3', 'N', 65.5, 105.0, 48.5], ['O2', 'C3', 'N', None, None, None], 
            ['C3', 'N', 'C1', None, None, None], ['C3', 'N', 'H5', None, None, None], 
            ['C1', 'N', 'H5', None, None, None]]
    dihedralparam=[['C1', 'N', 'C3', 'O1', 2, 0.4, 180.0, 68.0], 
            ['C2', 'O1', 'C3', 'N', 2, 1.24, 180.0, 48.5], 
            ['C2', 'O1', 'C3', 'N', 3, 2.27, 180.0, 48.5], 
            ['O1', 'C2', 'C1', 'N', 3, 0.2, 0.0, 32.0], 
            ['O1', 'C3', 'N', 'H5', 2, 1.27, 180.0, 68.0]]

    assert resn.bondparam == bondparam
    assert resn.angleparam == angleparam
    assert resn.dihedralparam == dihedralparam
    
    mol = Molecule() 
    mol.readmol2("mainaddi/oxaz.mol2")

    resn.loadGeometry(mol) 
    
    outfile = tmp_path + "/oxaz.pdb" 
    writepdb(resn,outfile)

    with open("mainaddi/oxaz_akit.pdb", 'rb') as file1:
        content1 = file1.read()

    with open(outfile, 'rb') as file2:
        content2 = file2.read()

    assert content1 == content2

def test_genbond_mol():
    from amolkit.molecule import Molecule
    mol=Molecule()
    mol.readxyz("readcoor/gluc.xyz")
    mol.genBonds()
    bonds = [[1, 2], [1, 7], [1, 11], [2, 1], [2, 3], [2, 8], [3, 2], [3, 4], [3, 9], 
            [4, 3], [4, 5], [4, 10], [5, 4], [5, 6], [5, 11], [6, 5], [6, 12], [7, 1], 
            [8, 2], [9, 3], [10, 4], [11, 1], [11, 5], [12, 6]]
    
    print(mol.bondindices) 
    assert mol.bondindices == bonds

def test_readcrd():
    from amolkit.molecule import Molecule
    mol=Molecule()
    mol.readcrd("readcoor/meoh.crd")
    print (mol.atomcoord_byIdx)
    atomcoord = OrderedDict([(1, [0.986, 0.015, 0.101]), (2, [2.402, 0.01, 0.091]), 
        (3, [0.625, 1.046, 0.065]), (4, [0.625, -0.478, 1.006]), 
        (5, [0.623, -0.526, -0.776]), (6, [2.694, 0.497, 0.88])])
    assert mol.atomcoord_byIdx == atomcoord

def test_readpdb():
    from amolkit.molecule import Molecule
    mol=Molecule()
    mol.readpdb("readcoor/meoh.pdb")
    atomcoord = OrderedDict([(1, [0.986, 0.015, 0.101]), (2, [2.402, 0.01, 0.091]), 
        (3, [0.625, 1.046, 0.065]), (4, [0.625, -0.478, 1.006]), 
        (5, [0.623, -0.526, -0.776]), (6, [2.694, 0.497, 0.88])])
    print (mol.atomcoord_byIdx)
    assert mol.atomcoord_byIdx == atomcoord

def test_writecrd(tmp_path):
    from amolkit.molecule import Molecule
    from amolkit.molecule import writecrd
    mol=Molecule()
    mol.readcrd("readcoor/meoh.crd")
    outfile = tmp_path + "/meoh.crd" 
    writecrd(mol,outfile)
    #writecrd(mol,tmp_path+"/meoh.crd")

    with open("readcoor/meoh_akit.crd", 'rb') as file1:
        content1 = file1.read()

    #with open(tmp_path"/meoh.crd", 'rb') as file2:
    with open(outfile, 'rb') as file2:
        content2 = file2.read()

    assert content1 == content2

def test_writepdb(tmp_path):
    from amolkit.molecule import Molecule
    from amolkit.molecule import writepdb
    mol=Molecule()
    mol.readpdb("readcoor/meoh.pdb")
    outfile = tmp_path + "/meoh.pdb" 
    writepdb(mol,outfile)
    #writepdb(mol,tmp_path"/meoh.pdb")

    with open("readcoor/meoh_akit.pdb", 'rb') as file1:
        content1 = file1.read()

    #with open(tmp_path"/meoh.pdb", 'rb') as file2:
    with open(outfile, 'rb') as file2:
        content2 = file2.read()

    assert content1 == content2

def test_writexyz(tmp_path):
    from amolkit.molecule import Molecule
    from amolkit.molecule import writexyz
    mol=Molecule()
    mol.readpdb("readcoor/meoh.pdb")
    outfile = tmp_path +"/meoh.xyz" 
    writexyz(mol,outfile) 
    #writexyz(mol,tmp_path"/meoh.xyz")

    with open("readcoor/meoh_akit.xyz", 'rb') as file1:
        content1 = file1.read()

    #with open(tmp_path"/meoh.xyz", 'rb') as file2:
    with open(outfile, 'rb') as file2:
        content2 = file2.read()

    assert content1 == content2

if __name__ == "__main__":
    tmp_path=sys.argv[1]
    test_readaddi()
    test_readdrud()
    test_genpsf_readaddi()
    test_genpsf_readdrud()
    test_genpsf_mainaddi()
    test_genpsf_maindrud()
    test_molsystem(tmp_path)
    test_genbond_mol()
    test_readcrd()
    test_readpdb()
    test_writecrd(tmp_path)
    test_writepdb(tmp_path)
    test_writexyz(tmp_path)




import os, sys
import numpy as np
from operator import sub
from collections import OrderedDict

class GetBondAngleDihedral():
    def __init__():
        """Initialize"""

    def GetBondDict(iatomname,bondlist,onlyatoms=True):
        """
        Generate Dict with atomname as keys and bonded atoms as values
        Arguments:
            iatomname: Dictionary of atom indices containing atomnames
            bondlist: List of list of bonds obtained from reading mol2 format
            onlyatoms: If true, output dictionary will not have lone pair names as keys
        Returns:
            aBond: Dictionary of atomnames with bonded atomnames as values
        """

        def patch_dict(anolpbond_dict,alpbond_dict):
            abond_dict=OrderedDict()
            for key,value in anolpbond_dict.items():
                abond_dict[key]=value
            if alpbond_dict != None:            
                for key,value in alpbond_dict.items():
                    chkval=anolpbond_dict.get(key)
                    if not chkval:
                        abond_dict.update({key:value})
                    else:
                        newvalue=chkval+value
                        abond_dict.update({key:newvalue})
            return (abond_dict)    

        anolpbond_dict=OrderedDict()
        alpbond_dict=OrderedDict()

        for field in bondlist:
            ba = iatomname[int(field[0])]
            bb = iatomname[int(field[-1])]
            if 'LP' not in [ba[0:2], bb[0:2]]:
                try:
                    anolpbond_dict[ba].append(bb)
                except KeyError:
                    anolpbond_dict[ba]=[bb]
                try:
                    anolpbond_dict[bb].append(ba)
                except KeyError:
                    anolpbond_dict[bb]=[ba]
            else:
                if onlyatoms:
                    if ba[0:2]=="LP":
                        lpa=ba
                        hatm=bb
                    else:
                        lpa=bb
                        hatm=ba
                    try:
                        alpbond_dict[hatm].append(lpa)
                    except KeyError:
                        alpbond_dict[hatm]=[lpa]
                else:
                    try:
                        alpbond_dict[ba].append(bb)
                    except KeyError:
                        alpbond_dict[ba]=[bb]
                    try:
                        alpbond_dict[bb].append(ba)
                    except KeyError:
                        alpbond_dict[bb]=[ba]

        if alpbond_dict != None:
            abond_dict=patch_dict(anolpbond_dict,alpbond_dict)
            return({"aBond":abond_dict,"aNoLPBond":anolpbond_dict,"aLPBond":alpbond_dict})
        else:
            abond_dict=anolpbond_dict
            return({"aBond":abond_dict})

    def GetBonds(aatomname,bond_dict):
        """
        Make a list of list of bond pairs with atomnames and atomindices
        Arguments:
            aatomname: Dictionary of atom names containing atomindices
            bond_dict: BondDict obtained from GetBondDict
        Returns:
            abondlist: list of list of bond pairs with atomnames
            ibondlist: list of list of bond pairs with atomindices
        """
        abondlist = []
        ibondlist = []
        for key,value in list(bond_dict.items()):
            for val in value:
                if [val,key] not in abondlist:
                    abondlist.append([key,val])
                    ibondlist.append([int(aatomname[key]),int(aatomname[val])])
        return(abondlist,ibondlist)

    def GetAngles(aatomname,bond_dict):
        """Make a list of list of angle triplets with atomnames and atomindices"""
        aanglelist = []
        ianglelist = []
        for k0,v0 in list(bond_dict.items()):
            if len(v0) > 1:
                for i in range(0,(len(v0)-1)):
                    ang2 = k0
                    ang1 = v0[i]
                    for j in range(i+1,len(v0)):
                        ang3 = v0[j]
                        aanglelist.append([ang1,ang2,ang3])
                        ianglelist.append([int(aatomname[ang1]),int(aatomname[ang2]),int(aatomname[ang3])])
        return(aanglelist,ianglelist)

    def GetDihedrals(aatomname,bond_dict):
        """Make a list of list of dihedral quads with atomnames and atomindices"""
        adihedrallist = []
        idihedrallist = []
        for k0,v0 in list(bond_dict.items()):
            dih0 = k0
            for k1 in v0:
                if bond_dict.get(k1) and len(bond_dict[k1]) != 1:
                    dih1 = k1
                    for k2 in bond_dict[k1]:
                        if bond_dict.get(k2) and len(bond_dict[k2]) != 1 and k2 != dih0:
                            dih2 = k2
                            for k3 in bond_dict[k2]:
                                if k3 != dih1 and k3 != dih0:
                                    dih3 = k3
                                    if [dih3,dih2,dih1,dih0] not in adihedrallist:
                                        adihedrallist.append([dih0,dih1,dih2,dih3])
                                        idihedrallist.append([int(aatomname[dih0]),int(aatomname[dih1]),int(aatomname[dih2]),int(aatomname[dih3])])
        return(adihedrallist,idihedrallist)

    def GetOnefives(aatomname,bond_dict):
        """Make a list of list of onefives 5 entries with atomnames and atomindices"""
        aonefivelist = []
        ionefivelist = []
        for k0,v0 in list(bond_dict.items()):
            skip=[k0,*v0]
            of0 = k0
            for k1 in v0:
                if len(bond_dict[k1]) != 1 and bond_dict.get(k1):
                    of1 = k1
                    for k2 in bond_dict[k1]:
                        if len(bond_dict[k2]) != 1 and k2 != of0:
                            of2 = k2
                            for k3 in bond_dict[k2]:
                                if len(bond_dict[k3]) != 1 and k3 != of1 and k3 != of0:
                                    of3 = k3
                                    for k4 in bond_dict[k3]:
                                        if k4 != of2 and k4 != of1 and k4 != of0:
                                            of4 = k4
                                            if [of4,of3,of2,of1,of0] not in aonefivelist:
                                                aonefivelist.append([of0,of1,of2,of3,of4])
                                                ionefivelist.append([int(aatomname[of0]),int(aatomname[of1]),int(aatomname[of2]),int(aatomname[of3]),int(aatomname[of4])])
        return(aonefivelist,ionefivelist)

    def GetImpropers(aatomname,bond_dict):
        aimplist=[]
        iimplist=[]
        for key,val in list(bond_dict.items()):
            if len(val)==3:
                aimplist.append([key,*val])
                iimplist.append([int(aatomname[key]),int(aatomname[val[0]]),int(aatomname[val[1]]),int(aatomname[val[2]])])
        return (aimplist,iimplist)

def generateBonds(elementnames=[],coordinates=[]):
    """
    Expects names and coordinates 
    """
    from amolkit import getEleInfo as gei 

    if elementnames and coordinates and len(elementnames) != len(coordinates):
        raise IndexError("names list and coordinate list don't match")
     
    coordinates = np.array(coordinates,dtype ='float32')

    delta = [None]*3
    for i in range(3):
       coord = coordinates[:,i]
       delta[i] = coord - coord.reshape((len(coord),1))

    dist = delta[0]**2 + delta[1]**2 + delta[2]**2 + 10. * np.identity(len(coordinates))

    radius = []
    for el in elementnames:
        radius.append(gei.covrad_by_atomsymbol(el)) 
    radius = np.array(radius, dtype='float32')

    radcov = (radius + radius.reshape((len(radius),1)))*2
    bonded = np.array(np.where(dist - radcov<0.)).transpose()
    distmat = np.sqrt(dist[bonded[:,0],bonded[:,1]])
    
    # amolkit indices for molecules starts from 1
    bondindices=[]
    for b in bonded:
        bondindices.append(list(map(lambda x:x+1,b)))

    bondinfo={}
    bondinfo["radcov"]         = radcov
    bondinfo["pybondindices"]  = bonded
    bondinfo["distmat"]        = distmat
    bondinfo["bondindices"]    = bondindices
    return bondinfo 

def getBondValue(coord1,coord2):
    coord1=np.array(coord1,dtype='float32')
    coord2=np.array(coord2,dtype='float32')

    xyz = np.array(list(map(sub,coord1,coord2)))
    distance = np.linalg.norm(xyz)
    distance = round(distance,4)
    return distance

def getAngleValue(coord1,coord2,coord3):
    coord1=np.array(coord1,dtype='float32')
    coord2=np.array(coord2,dtype='float32')
    coord3=np.array(coord3,dtype='float32')

    bav = coord1 - coord2 
    bcv = coord3 - coord2 
    cosine_angle = np.dot(bav, bcv) / (np.linalg.norm(bav) * np.linalg.norm(bcv))
    angle = np.arccos(cosine_angle)
    angle = np.degrees(angle)
    angle = round(angle,3)
    return angle

def getDihedralValue(coord1,coord2,coord3,coord4):
    coord1=np.array(coord1,dtype='float32')
    coord2=np.array(coord2,dtype='float32')
    coord3=np.array(coord3,dtype='float32')
    coord4=np.array(coord4,dtype='float32')
    b0 = -1.0*(coord2 - coord1)
    b1 = coord3 - coord2 
    b2 = coord4 - coord3 
    b0xb1 = np.cross(b0, b1)
    b1xb2 = np.cross(b2, b1)
    b0xb1_x_b1xb2 = np.cross(b0xb1, b1xb2)
    y = np.dot(b0xb1_x_b1xb2, b1)*(1.0/np.linalg.norm(b1))
    x = np.dot(b0xb1, b1xb2)
    dihedral = np.arctan2(y, x)
    dihedral = np.degrees(dihedral)
    dihedral = round(dihedral,3)
    return dihedral

def getICvalue(coord1=[],coord2=[],coord3=[],coord4=[]):
    if coord1 and coord2 and not coord3 and not coord4: return getBondValue(coord1,coord2)
    if coord1 and coord2 and coord3 and not coord4: return getAngleValue(coord1,coord2,coord3)
    if coord1 and coord2 and coord3 and coord4: return getDihedralValue(coord1,coord2,coord3,coord4)



import sys
import math
import argparse
from collections import OrderedDict
from amolkit import genic
from amolkit import getEleInfo as gei 
from amolkit.stringmanip import getPenaltyfromString

def readCharmmTopology(resname,resitopfile,bio=True):
    '''
    Load the features of the residue based on topology file.

    Note::
    field[1] is atomname
    field[2] is atomtype
    if bio is True then utilize gei.atomicmass_by_anyatomname upon atomtype
    to get atomicmass or symbol. This way I am not limited to predefined atomtypes in EleInfo.

    atommass_byId[field[1]] = gei.atomicmass_by_anyatomname(field[2],bio)
    atomele_byId[field[1]] = gei.atomsymbol_by_anyatomname(field[2],bio)

    But if it false, then I should ideally get atomicmass by atomtype
    and in case an atomtype is missing, I should allow user to supply new atomtype and its correspondence.

    # Give some thought if bio is False. In that case, atomtype = MAG will not be converted to Mg
    # Thus it would not find Mg
    '''

    topology = {}
    topology["resname"]            = resname
    topology["fftype"]             = 'additive' 
    topology["rescharge"]          = None 
    topology['ntopatoms']          = None
    topology['natoms']             = None
    topology['first_byId']         = None
    topology['last_byId']          = None
    topology['atomindex_byId']     = OrderedDict() 
    topology['atomname_byIdx']     = OrderedDict()
    topology['atomtype_byId']      = OrderedDict()
    topology['atomcharge_byId']    = OrderedDict()
    topology['atompenalty_byId']   = OrderedDict()
    topology['atommass_byId']      = OrderedDict()
    topology['atomele_byId']       = OrderedDict()
    topology['atomalpha_byId']     = OrderedDict()
    topology['atomthole_byId']     = OrderedDict()
    topology['atomdrudetype_byId'] = OrderedDict()
    topology['atomsinring']        = OrderedDict()
    topology['groups']             = []  
    topology['atomnames']          = []    
    topology['bonds']              = []
    topology['lpbonds']            = []
    topology['angles']             = []
    topology['dihedrals']          = []
    topology['impropers']          = []
    topology['lpics']              = []
    topology['anisotropies']       = []
    topology['donors']             = []
    topology['acceptors']          = []
    topology['cmaps']              = []
    topology['atomcharges']        = []

    try:
        filein=open(resitopfile, "r")
    except IOError:
        raise IOError('Could not access the ' + resitopfile + ' topology file')

    found=False
    ring = 0
    nat = 0

    for nl,line in enumerate(filein):
        #Strip the line, remove the comment lines, and split using whitespace
        field = line.strip().split("!")[0].split()
        
        fieldup = list(map(lambda x:x.upper(),field))

        comment = "!".join(line.strip().split("!")[1:]) 

        if len(fieldup) >= 1:
            ftype = fieldup[0][0:4]
            if found and ftype == "END": break
            if ftype in ["RESI","PRES"] and fieldup[1] !=  topology['resname'].upper():
                if found:
                    break
                else:
                   found = False
            elif ftype in ["RESI","PRES"] and fieldup[1] ==  topology['resname'].upper():
                found = True
                try:
                    topology["rescharge"]=round(float(field[2]),3)
                except (ValueError,IndexError):
                    raise ValueError('Beware:: Error in detecting residue charge of '+topology['resname']+'.')

        if found and len(fieldup) > 0:
            ftype = fieldup[0][0:4]
            if ftype not in ["RING","GROU","ATOM","BOND","DOUB","TRIP","ANGL","DIHE","IMPR","DONO","ACCE","CMAP","LONE","ANIS"]: continue
            if ftype == "RING":
                ring=ring+1
                #topology['atomsinring'][ring]=field[1:]
                topology['atomsinring'][ring]=fieldup[1:]
                continue
            elif ftype == "GROU":
                topology['groups'].append([nat,0,0])
                continue
            elif ftype == "ATOM" and len(fieldup) > 3: 
                nat = nat + 1
                topology['atomindex_byId'][fieldup[1]] = nat
                topology['atomname_byIdx'][nat] = fieldup[1]    # Atomname stored as in str
                topology['atomtype_byId'][fieldup[1]] = fieldup[2]  # Atomtype capitalized before storing
                topology['atomcharge_byId'][fieldup[1]] = float(field[3])
                # Using atomicmass_by_anyatomname, but supplying atomtype
                # Because atomicmass_by_anyatomtype does a fixed conversion
                topology['atommass_byId'][fieldup[1]] = gei.atomicmass_by_anyatomname(field[2],bio)
                topology['atomele_byId'][fieldup[1]] = gei.atomsymbol_by_anyatomname(field[2],bio)
                topology['atomcharges'].append(float(field[3]))
                topology['atompenalty_byId'][fieldup[1]],_ = getPenaltyfromString(comment)

                if topology['atomnames'] and fieldup[1] in topology['atomnames']:
                    raise ValueError("Topology contains repeated atomname %s. Unable to load."%(field[1]))
                else:
                    topology['atomnames'].append(fieldup[1]) 
                
                if any(list(map(lambda x:x in fieldup, ["ALPHA","THOLE","TYPE"]))): 
                    topology['atomalpha_byId'][fieldup[1]]=None
                    topology['atomthole_byId'][fieldup[1]]=None
                    topology['atomdrudetype_byId'][fieldup[1]]='DRUD'
                    topology["fftype"] = "drude"

                if "ALPHA" in fieldup:
                    topology['atomalpha_byId'][fieldup[1]]=float(field[fieldup.index("ALPHA")+1])
                if "THOLE" in fieldup:
                    topology['atomthole_byId'][fieldup[1]]=float(field[fieldup.index("THOLE")+1])
                if "TYPE" in fieldup:
                    topology['atomdrudetype_byId'][fieldup[1]]=fieldup[fieldup.index("TYPE")+1]

                continue

            #Bond list dictionary of each atom is populated.
            elif ftype in ["BOND","DOUB","TRIP"] and (len(fieldup)-1) % 2 == 0:
                for i in range(1,len(fieldup),2):
                    ba = fieldup[i].strip("+-")
                    bb = fieldup[i+1].strip("+-")
                    topology['bonds'].append([ba,bb])

                    if ba.startswith('+'): topology['first_byId'] = ba
                    if bb.startswith('+'): topology['first_byId'] = bb
                    if ba.startswith('-'): topology['last_byId'] = ba
                    if bb.startswith('-'): topology['last_byId'] = bb

                    if 'LP' in [ba[0:2], bb[0:2]]:
                        topology['lpbonds'].append([ba,bb])

            elif ftype in ["BOND","DOUB","TRIP"] and (len(fieldup)-1) % 2 != 0:
                 raise ValueError('Topology file BOND line has the wrong format. It should be multiples of two! BOND At1 At2 At2 At3 At1 At3 etc.')

            #Angle list dictionary.
            elif ftype == "ANGL" and (len(fieldup)-1) % 3 == 0:
                for i in range(1,len(fieldup),3):
                    aa = fieldup[i].strip("+-")
                    ab = fieldup[i+1].strip("+-")
                    ac = fieldup[i+2].strip("+-")
                    topology['angles'].append([aa,ab,ac])
                continue
            elif ftype == "ANGL" and (len(fieldup)-1) % 3 != 0:
                raise ValueError('Topology file ANGLE line has the wrong format. Ignored reading.')
                continue

            #Dihedral list dictionary.
            elif ftype == "DIHE" and (len(fieldup)-1) % 4 == 0:
                for i in range(1,len(fieldup),4):
                    da = fieldup[i].strip("+-")
                    db = fieldup[i+1].strip("+-")
                    dc = fieldup[i+2].strip("+-")
                    dd = fieldup[i+3].strip("+-")
                    topology['dihedrals'].append([da,db,dc,dd])

            elif ftype == "DIHE" and (len(fieldup)-1) % 4 != 0:
                raise ValueError('Topology file DIHEDRAL line has the wrong format. Ignored reading.')

            #Improper list dictionary.
            elif ftype == "IMPR" and (len(fieldup)-1) % 4 == 0:
                for i in range(1,len(fieldup),4):
                    ima = fieldup[i].strip("+-")
                    imb = fieldup[i+1].strip("+-")
                    imc = fieldup[i+2].strip("+-")
                    imd = fieldup[i+3].strip("+-")
                    topology['impropers'].append([ima,imb,imc,imd])
            elif ftype == "IMPR" and (len(fieldup)-1) % 4 != 0:
                raise ValueError('Topology file IMPROPER line has the wrong format. Ignored reading.')

            #Donor list dictionary.
            elif ftype == "DONO" and (len(fieldup)-1) % 2 == 0:
                for i in range(1,len(fieldup),2):
                    topology['donors'].append([fieldup[i+1],fieldup[i]])

            elif ftype == "DONO" and (len(fieldup)-1) % 2 != 0:
                raise ValueError('Topology file DONOR line has the wrong format. Ignored reading.')

            #Acceptor list dictionary.
            elif ftype == "ACCE" and (len(fieldup)-1) % 2 == 0:
                for i in range(1,len(fieldup),2):
                    topology['acceptors'].append([fieldup[i],fieldup[i+1]])

            elif ftype == "ACCE" and (len(fieldup)-1) == 1:
                topology['acceptors'].append([fieldup[-1],""])

            elif ftype == "ACCE" and (len(fieldup)-1) != 1 or ftype == "ACCE" and (len(fieldup)-1) % 2 != 0:
                raise ValueError('Topology file ACCEPTOR line has the wrong format. Ignored reading.')

            elif ftype == "CMAP" and (len(fieldup)-1) % 8 == 0:
                for i in range(1,len(fieldup),8):
                    ca = fieldup[i].strip("+-")
                    cb = fieldup[i+1].strip("+-")
                    cc = fieldup[i+2].strip("+-")
                    cd = fieldup[i+3].strip("+-")
                    ce = fieldup[i+4].strip("+-")
                    cf = fieldup[i+5].strip("+-")
                    cg = fieldup[i+6].strip("+-")
                    ch = fieldup[i+7].strip("+-")
                    topology['cmaps'].append([ca,cb,cc,cd,ce,cf,cg,ch])

            elif ftype == "CMAP" and (len(fieldup)-1) % 8 != 0:
                raise ValueError('Topology file CMAP line has the wrong format. Ignored reading.')
            
            # topology['atomindex_byId'] will change later according to inclusion of 
            # Lone pair list dictionary
            elif ftype == "LONE": 
                if fieldup[1][0:4] in ["RELA","BISE"]:
                    try:
                        hlist = fieldup[2:6]
                        vlist = []
                        for ilp in range(len(fieldup)): 
                            if fieldup[ilp][0:4] == "DIST" and fieldup[1][0:4] == "RELA":
                                vlist.append(float(fieldup[ilp+1]))
                            if fieldup[ilp][0:4] == "DIST" and fieldup[1][0:4] == "BISE":
                                vlist.append(-(float(fieldup[ilp+1])))
                            if fieldup[ilp][0:4] == "ANGL":
                                vlist.append(float(fieldup[ilp+1]))
                            if fieldup[ilp][0:4] == "DIHE":
                                vlist.append(float(fieldup[ilp+1]))
                        topology['lpics'].append([fieldup[1][0:4],hlist,vlist])
                    except (IndexError, ValueError,KeyError):    
                        raise ValueError('Topology file LONEPAIR line has the wrong format. Ignored reading.')
                elif fieldup[1][0:4] == "COLI": 
                    try:
                        hlist =fieldup[2:5]
                        vlist = []
                        for ilp in range(len(fieldup)): 
                            if fieldup[ilp][0:4] == "DIST":
                                vlist.append(float(fieldup[ilp+1]))
                            if fieldup[ilp][0:4] == "SCAL":
                                vlist.append(float(fieldup[ilp+1]))
                        vlist.append(float(0.0))
                        topology['lpics'].append([fieldup[1][0:4],hlist,vlist])
                    except (IndexError, ValueError):    
                        raise ValueError('Topology file LONEPAIR line has the wrong format. Ignored reading.')
                elif fieldup[1][0:4] == "FIXE":
                    try:
                        topology['lpics'].append([[fieldup[1][0:4]],[],list(map(float,fieldup[2:5]))]) 
                    except (IndexError, ValueError):    
                        raise ValueError('Topology file LONEPAIR line has the wrong format. Ignored reading.')
                elif fieldup[1][0:4] == "CENT":
                    try:
                        topology['lpics'].append([fieldup[1][0:4],fieldup[2:],[]])
                    except (IndexError, ValueError):    
                        raise ValueError('Topology file LONEPAIR line has the wrong format. Ignored reading.')

            # Anisotropy list dictionary
            elif ftype == "ANIS": 
                try:
                    hlist = field[1:5]
                    for ilp in range(len(fieldup)): 
                        if fieldup[ilp][0:3] == "A11":
                            a11 =  float(fieldup[ilp+1])
                        if fieldup[ilp][0:3] == "A22":
                            a22 =  float(fieldup[ilp+1])
                    if a11 == 0.0 or a22 == 0.0:
                       logging.error('0.0 is invalid force constant. Ignoring this anisotropy line.')
                       continue
                    vlist = [a11,a22] 
                    topology['anisotropies'].append([hlist,vlist])
                except (IndexError, ValueError,KeyError):    
                    raise ValueError('Topology file Anisotropy line has the wrong format. Ignored reading.')
    #### Reading done ###

    if found:
        onlyatomsnlps = [at for at in topology['atomindex_byId'].keys() if at[0].upper() != "D"]
        onlyatoms     = [at for at in onlyatomsnlps if at[0:2].upper() != "LP" ]
        topology['ntopatoms']     = len(topology['atomindex_byId'].keys())
        topology['natoms']        = len(onlyatoms)

        #Check for residue total charge
        if abs( sum(topology['atomcharges']) - topology['rescharge'] ) > 0.00001:
            print('Warning: Your total charge does not match the sum of partial charges.')

    elif not found:
        print ('Could not find the residue name: '+topology['resname']+' in the provided topology file.')

    return topology


class Topology():
    def __init__(self): 
        """
        Constructor for the Topology class.
        """

    def loadCharmmTopology(self,resname,resitopfile): 
        """
        Load topology information using CHARMM force field and store it as attributes.
        
        Args:
            resname (str): Name of the residue.
            resitopfile (str): Path to the CHARMM topology file.
        """

        topology = readCharmmTopology(resname,resitopfile)
        for key,value in topology.items():
            self._assignattr(key,value)
    
    def _assignattr(self, key, value):
        """
        Private method to assign attributes to the class instance.

        Args:
            key: The attribute name.
            value: The value to assign to the attribute.
        """
        setattr(self, key, value)  




import numpy as np
from amolkit import EleInfo as ei
from amolkit.stringmanip import stripname


cpatomlist=list(map(lambda x:x.upper(), ei.AtomOrder))
cpbiolist=list(ei.BioNametoAtomName.keys())
elementname = {v: k for k, v in ei.AtomicNumber.items()}

def bioname_to_atomname(atomname):
    bioname=ei.BioNametoAtomName.get(atomname.upper())
    return bioname

def getindex(atomlist:list,atomname):
    index=None
    try:
       index=atomlist.index(atomname[0:3].strip())
       atomname=atomname[0:3]
    except ValueError:
       try:
          index=atomlist.index(atomname[0:2].strip())
          atomname=atomname[0:2]
       except ValueError:
          try:
              index=atomlist.index(atomname[0:1].strip())
              atomname=atomname[0:1]
          except ValueError:
              index=None
              atomname=None
    return index,atomname

def atomsymbol_by_anyatomname(atomname,bio=True):
    inname=atomname
    atomname=stripname(atomname) 
    atomname=atomname.upper()
    if bio: 
        bioindex,bioname=getindex(cpbiolist,atomname)
        if bioindex:        
            atomname=ei.BioNametoAtomName.get(bioname)
            return atomname
    
    index,name=getindex(cpatomlist,atomname)
    atomname=ei.AtomOrder[index] if index else name
    if not atomname: raise NameError('Could not determine correct atomsymbol for %s'%(inname))
    return atomname

def atomicnumber_by_anyatomname(atomname,bio=True):
    atomsymbol=atomsymbol_by_anyatomname(atomname,bio)
    atomicnumber=ei.AtomicNumber.get(atomsymbol)
    return atomicnumber

def atomicmass_by_anyatomname(atomname,bio=True):
    atomsymbol=atomsymbol_by_anyatomname(atomname,bio)
    atomicmass=ei.AtomicMass.get(atomsymbol)
    return atomicmass

def atomicnumber_by_atomtype(atomtype):
    eleName=ei.AtomTypetoAtomName.get(atomtype.upper())
    if eleName:
       eleNumber=ei.AtomicNumber[eleName]
       return eleNumber
    else:  
       raise NameError("Could not find any element associated with " + atomtype) 

def atomicmass_by_atomtype(atomtype):
    eleName=ei.AtomTypetoAtomName.get(atomtype.upper())
    if eleName:
       eleMass=ei.AtomicMass[eleName]
       return (eleMass)
    else:
       raise NameError("Could not find any element associated with " + atomtype) 

def covrad_by_atomsymbol(atomname): 
    radius=ei.CovalentRadius[atomname]
    return radius

def color_by_atomsymbol(atomname): 
    color=ei.AtomColor[atomname]
    return color

def atomsymbol_by_atomicnumber(atomicnumber): #:int):
    atomsymbol=elementname[int(atomicnumber)]
    return atomsymbol

class ElementProp():
   def __init__(self, Z, radius, max_bonds, r, g, b):
      self.Z = Z
      self.radius = radius
      self.max_bonds = max_bonds
      self.color = np.array([r,g,b], dtype='float32')
   def __str__(self):
      format_string = ' '.join(['{:4d}','{:9f}','{:4d}','{:9f}'*3])
      return format_string.format(self.Z, self.radius, self.max_bonds, *self.color)

class AssignElementProp(ElementProp):
   def __init__(self): 
      self.PSE = {}
      self.ATSYM = {}
      for atsym in ei.AtomOrder:
         self.PSE[atsym.upper()] = ElementProp(ei.AtomicNumber[atsym],ei.CovalentRadius[atsym],ei.BondOrder[atsym],*ei.AtomColor[atsym])
         self.ATSYM[ei.AtomicNumber[atsym]] = atsym



import math
from collections import OrderedDict
from amolkit import genic
from amolkit import getEleInfo as gei 
from amolkit.topology import Topology
from amolkit.molecule import Molecule

class MolSystem(Topology):
    """
    Create molecule object using topology, parameters and coordinate file
    """
    def __init__(self): 
        """
        """
        super().__init__()

    def loadTopology(self,resname,resitopfile):
        '''
        Load the features of the residue based on topology file.
        '''

        self.loadCharmmTopology(resname,resitopfile)

    def loadParameters(self,parfiles,autogenangdih=True):
        """
        Load parameters into molecule's internal coordinates by reading parameter file/s.
        Arguments:
            parfiles: Can be a list of parameter containing files or single file.
        Attributes:
            self.bondparam: List of parameters with bond ICs [[C1,H1,kval,eqdist,penalty],...]
            self.angleparam: List of parameters with angle ICs [[H1,C1,C2,kval,eqangle,penalty],...]
            self.dihedralparam: List of parameters with dihedral ICs [[C1,C2,C3,C4,mult,kval,phase,penalty],...]
                                Note that position of multiplicity has changed compared to charmm parameter format
                                This helps in management of the list. 
            self.superdihparam: Each dihedral IC has 6 entries with 6 different multiplicties.
                                If there is no dihedral value corresponding to a particular multiplicity in
                                charmm parameter file, it will just store None against it.
                                This helps when one wants to optimize parameters.
        Note:
            Have to add improper also. Haven't added because I dont have function to generate impropers in a molecule.
            Thus have to rely on impropers in topology file. self.impropers are stored in terms of numbers.
        """

        from copy import deepcopy
        from amolkit.parameter import Parameter

        param = Parameter()
        param.loadCharmmParameter(parfiles)
        #param.charmmParameter: Collection of all parameters obtained from parfiles.
        
        if not self.bonds:
            print ("Charmm Parameters obtained, but not loaded on molecular ics. You should load topology prior to loading parameters")
            return
        
        # Get list of bond angle and dihedral
        bondindices=[]
        for ba,bb in self.bonds:
            bondindices.append([self.atomindex_byId[ba],self.atomindex_byId[bb]])
        bondgraph = genic.GetBondAngleDihedral.GetBondDict(self.atomname_byIdx,bondindices)["aNoLPBond"]
        bondnames,_ = genic.GetBondAngleDihedral.GetBonds(self.atomindex_byId,bondgraph)
        if autogenangdih:
            anglenames,_ = genic.GetBondAngleDihedral.GetAngles(self.atomindex_byId,bondgraph)
            dihedralnames,_ = genic.GetBondAngleDihedral.GetDihedrals(self.atomindex_byId,bondgraph)
        else:
            anglenames = deepcopy(self.angles)
            dihedralnames = deepcopy(self.dihedrals)
            
        # Assign parameters to bond, angle and dihedral
        # Initialize parameter lists
        self.bondparam = deepcopy(bondnames)
        self.angleparam = deepcopy(anglenames)
        self.superdihparam = [None]*6*(len(dihedralnames))

        for i,bp in enumerate(self.bondparam):
            self.bondparam[i].append(None)
            self.bondparam[i].append(None) 
            self.bondparam[i].append(None) 

        # Will not assume Urey Bradley terms here, should be appended if exists
        for i,ap in enumerate(self.angleparam):
            self.angleparam[i].append(None)   
            self.angleparam[i].append(None) 
            self.angleparam[i].append(None) 
         
        # Each dihedral will have 6 copies with 6 different multiplicities 
        nd = 0    
        for i,dp in enumerate(dihedralnames):
            for j in range(6):
                self.superdihparam[nd] = deepcopy(dp) 
                self.superdihparam[nd].append(j+1) 
                self.superdihparam[nd].append(None) 
                self.superdihparam[nd].append(None) 
                self.superdihparam[nd].append(None) 
                nd=nd+1
                
        # Assign now
        for i,bl in enumerate(bondnames):
            parainq = [self.atomtype_byId[bl[0]], self.atomtype_byId[bl[1]]]
            for j, listele in enumerate(param.charmmParameter['BONDS']):
               listf = listele[0:2]
               listb = listele[0:2][::-1]
               if parainq == listf or parainq == listb:
                   self.bondparam[i][2] = listele[2] 
                   self.bondparam[i][3] = listele[3]
                   self.bondparam[i][4] = listele[4]

        for i,al in enumerate(anglenames):
            parainq = [self.atomtype_byId[al[0]], self.atomtype_byId[al[1]],self.atomtype_byId[al[2]]]
            for j, listele in enumerate(param.charmmParameter['ANGLES']):
               listf = listele[0:3]
               listb = listele[0:3][::-1]
               if parainq == listf or parainq == listb:
                   if len(listele) == 7:
                       self.angleparam[i][3] = listele[3]
                       self.angleparam[i][4] = listele[4]
                       self.angleparam[i][5] = listele[5]
                   elif len(listele) == 9:
                       self.angleparam[i][3] = listele[3]
                       self.angleparam[i][4] = listele[4]
                       self.angleparam[i][5] = listele[5]
                       self.angleparam[i].append(listele[6])
                       self.angleparam[i].append(listele[7])

        for i,dl in enumerate(dihedralnames):
            parainq = [self.atomtype_byId[dl[0]], self.atomtype_byId[dl[1]],self.atomtype_byId[dl[2]],self.atomtype_byId[dl[3]]]
            for j, listele in enumerate(param.charmmParameter['DIHEDRALS']):
               listf = listele[0:4]
               listb = listele[0:4][::-1]
               if parainq == listf or parainq == listb:
                   #listele[5] is multiplicity
                   index = i*6 + (int(listele[5])-1)
                   self.superdihparam[index][5] = listele[4] 
                   self.superdihparam[index][6] = listele[6] 
                   self.superdihparam[index][7] = listele[7] 
        
        # Get dihedral IC with same multiplicities as obtained in charmm parameter files
        # ==> But this also removes those entries for which no parameter is found.
        self.dihedralparam=[]
        for ent in self.superdihparam:
            if ent[5] != None:
                self.dihedralparam.append(ent)

    def _checkCompatibility(self):
        # sanity check
        namecompatible=False
        ordercompatible=False
        ordercompatible_nolpdx=False

        topnames = list(map(lambda x:x.upper(),self.atomnames)) 
        molnames = list(map(lambda x:x.upper(),self.molecule.atomname_byIdx.values()))

        # Ideal scenario 
        # All atomnames in topology and molecule instance are same and in order.
        if topnames == molnames:
            namecompatible = True

        if not namecompatible:    
            topnames = sorted(topnames)
            molnames = sorted(molnames)
            if topnames == molnames:
               namecompatible = True
        if not namecompatible:       
            topnames_nolpd = [v for v in topnames if v[0:2].upper() != "LP" and v[0].upper() != "D"]
            molnames_nolpd = [v for v in molnames if v[0:2].upper() != "LP" and v[0].upper() != "D"]
            if topnames_nolpd == molnames_nolpd:
                namecompatible = True
        if not namecompatible:    
            topnames_nolpd = sorted(topnames_nolpd)
            molnames_nolpd = sorted(molnames_nolpd)
            if topnames_nolpd == molnames_nolpd:
               namecompatible = True

        if not namecompatible:    
            topelements = list(map(lambda x:x.upper(),self.atomele_byId.values())) 
            molelements = list(map(lambda x:x.upper(),self.molecule.atomele_byIdx.values()))
            if topelements == molelements:
                ordercompatible = True

        if not namecompatible and not ordercompatible:    
            topelements_nolpdx = [v for v in topelements if v[0:2].upper() != "LP" and v[0].upper() != "D" and v[0].upper() != "X"]
            molelements_nolpdx = [v for v in molelements if v[0:2].upper() != "LP" and v[0].upper() != "D" and v[0].upper() != "X"]

            if topelements_nolpdx == molelements_nolpdx:
                ordercompatible_nolpdx = True
        return (namecompatible,ordercompatible,ordercompatible_nolpdx)        

    def loadGeometry(self,molecule,update=""):
        """
        This method expects to receive Molecule object with preloaded coordinate file.
        It uses the coordinates to assign atomcoord to molsystem object.
        Before assigning atomcoord, it checks if molecule coordinates are compatible with
        atomnames in topology object. Else raises ValueError

        Arguments:
            molecule: Molecule object with preloaded coordinate file
        Attributes:
            self.atomcoord: OrderedDict containing coordinates with respect atom name in topology object
        """

        self.molecule = molecule 
        self.nmolatoms = self.molecule.nmolatoms
        self.atomcoord_byId  = OrderedDict() 
        self.atomcoord_byIdx = OrderedDict() 
        self.atomele_byIdx   = OrderedDict() 
        self.atomserial_byIdx= OrderedDict() 
        self.atomresn_byIdx  = OrderedDict() 
        self.atomresid_byIdx = OrderedDict() 
        self.atomsegn_byIdx  = OrderedDict() 
        self.atomchain_byIdx = OrderedDict()
        self.atomoccu_byIdx  = OrderedDict()
        self.atomtfac_byIdx  = OrderedDict()
        self.atomcharge_byIdx= OrderedDict()

        namecompatible,ordercompatible,ordercompatible_nolpdx = self._checkCompatibility() # sanity check

        if namecompatible: 
            for key,value in self.atomindex_byId.items():
                try:
                    self.atomcoord_byId[key] = self.molecule.atomcoord_byIdx[self.molecule.atomindex_byId[key]]
                except KeyError:
                    self.atomcoord_byId[key]=[9999.999,9999.999,9999.999]
        
        elif ordercompatible:    
            for key,value in self.atomname_byIdx.items():
                try:
                    self.atomcoord_byId[value] = self.molecule.atomcoord_byIdx[key]
                except KeyError:
                    self.atomcoord_byId[value] = [9999.999,9999.999,9999.999]

        elif ordercompatible_nolpdx:    
            key = 0
            atomcoord_byIdx_nolpdx=OrderedDict() 
            for i in range(self.molecule.nmolatoms):
                checkatom = molecule.atomele_byIdx[i+1].upper() 
                if checkatom[0:2] == "LP" or checkatom[0] == "D" or checkatom[0] =="X": 
                    continue 
                else:
                    key = key + 1
                    atomcoord_byIdx_nolpdx[key] = self.molecule.atomcoord_byIdx[i+1] 

            key = 0
            for i in range(self.ntopatoms):
                value = self.atomname_byIdx[i+1]
                if value[0:2] == "LP" or value[0] == "D" or value[0] =="X": 
                    self.atomcoord_byId[value] = [9999.999,9999.999,9999.999]
                else:
                    key = key + 1
                    self.atomcoord_byId[value] = atomcoord_byIdx_nolpdx[key] 

        else:
            raise ValueError ("Neither the name nor the order of elements in molecule are compatible with topology")

        for key,value in self.atomcoord_byId.items(): 
            self.atomcoord_byIdx[self.atomindex_byId[key]]= value 
            self.atomserial_byIdx[self.atomindex_byId[key]]= self.atomindex_byId[key]
            self.atomele_byIdx[self.atomindex_byId[key]]= self.atomele_byId[key]
            self.atomresn_byIdx[self.atomindex_byId[key]]= self.resname 
            self.atomresid_byIdx[self.atomindex_byId[key]]= 1 
            self.atomsegn_byIdx[self.atomindex_byId[key]]=self.resname 

class genRDMolSmiles():
    '''
    This class is to create an RDMol object from the topology file. To generate initial coordinates for the molecule, you can run coordinates_fromrdmol , after you initialize the class.
    '''

    import numpy as np
    try:
        from rdkit import Chem
        from rdkit.Chem import AllChem
    except ImportError:
        logging.warning('RDKit not found!')

    def __init__(self,resi='resn',bonds=[],atomnames=[],smiles=None):
        self.resname=resi
        if isinstance(smiles,str):
            self.__rdmol_smiles(smiles)
        else:
            raise ValueError('Either provide a bonds and atomnames list&dictionary or pass a smiles string.')

    def __rdmol_smiles(self,smiles):
        m=Chem.MolFromSmiles(smiles)
        try:
            Chem.SanitizeMol(m)
        except ValueError:
            pass
        self.mol=m
        #coordinates_fromrdmol()

    def coordinates_fromrdmol(self,outpath,cortype='crd'):
        AllChem.EmbedMolecule(m,useExpTorsionAnglePrefs=True,useBasicKnowledge=True)
        AllChem.MMFFOptimizeMolecule(m)

        #names not used but it is returned from this function by default
        names,positions=mr.Molecule.readpdb(Chem.rdmolfiles.MolToPDBBlock(m).split('\n'))
        self.coor=positions
        with open(fileout,'w') as f:
            if otype == 'crd':
                f.write("*Created by FFParam\n*\n")
                f.write("%10i  EXT\n" %len(self.rdmolnrs))
            for i in self.rdmolnrs:
                name=self.rdmolnrs[i]
                xyz=list(positions[i])
                nr=i+1
                line=cw.crdline([nr,name]+xyz,resname=self.resname,segname=self.resname,returntype=otype)
                f.write(line+'\n')
        return(fileout)

        if otype == 'mol2':
            mol2=self.rdmoltomol2(m)
            with open(fileout,'w') as f:
                f.write(mol2)
            return(fileout)
        else:
            logging.error('Other types from rdkit are not yet supported.')
            return(None)

    def rdmoltomol2(self,m):
       # #Create a moltomol2 function that assigns unique names to each atom and embeds the residue name
        pMol2Block =""
        Chem.Kekulize(m)
        mH =Chem.AddHs(m)
        if mH != m:
            m=mH
        m.SetProp("_Name",self.resname)
        try:
           AllChem.EmbedMolecule(m,useExpTorsionAnglePrefs=True,useBasicKnowledge=True)
        except ValueError:
           logging.error("Cant create coordinates, hypervalency exists")
           return
        AllChem.MMFFOptimizeMolecule(m)
        natoms=m.GetNumAtoms()
        nbonds=m.GetNumBonds()
        names,positions=mr.Molecule.readpdb(Chem.rdmolfiles.MolToPDBBlock(m).split('\n'))
        pMol2Block = "@<TRIPOS>MOLECULE\n"
        pMol2Block += "%s\n"%(m.GetProp("_Name"))
        pMol2Block += " %d %d %d %d %d\n"%(natoms,nbonds,0,0,0)
        pMol2Block += "SMALL\n"
        pMol2Block += "GASTEIGER\n"
        pMol2Block += "\n"
        pMol2Block += "@<TRIPOS>ATOM\n"
        for i, atom in enumerate(m.GetAtoms()):
            atname = atom.GetSymbol()+str(i+1)
            attype = atom.GetSymbol()
            parchar= round(float(0),4)
            xyz=list(positions[i])
            pMol2Block += "%7d %7s %10.4f%10.4f%10.4f %5s%4s  %8s%10.4f\n"%(i+1,atname,xyz[0],xyz[1],xyz[2],attype,"1",self.resname.upper(),parchar)
        pMol2Block += "@<TRIPOS>BOND\n"
        for i, bi in enumerate(m.GetBonds()):
             i1 = int(bi.GetBeginAtomIdx())
             i2 = int(bi.GetEndAtomIdx())
             bint = bi.GetBondType()
             btype = "1"
             if(bint==1 or bint == 'SINGLE'): btype = "1"
             if(bint==2 or bint == 'DOUBLE'): btype = "2"
             if(bint==12 or bint == 'AROMATIC'): btype = "ar"
             if(bint==3 or bint == 'TRIPLE'):  btype = "3"
             pMol2Block += "%6d%6d%6d %s\n"%(i+1,i1+1,i2+1,btype)
        return pMol2Block
